@@ -1,10 +1,8 @@
 #!/usr/bin/perl
 #
-# $HeadURL: https://svn.spodhuis.org/ksvn/sieve-connect/sieve-connect.pl $
+# MANAGESIEVE (timsieved) client script
 #
-# timsieved client script
-#
-# Copyright © 2006-2012 Phil Pennock.  All rights reserved.
+# Copyright © 2006-2013 Phil Pennock.  All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions
@@ -46,27 +44,40 @@ my $DEFAULT_PORT = 'sieve(4190)';
 my %ssl_options = (
 	SSL_version	=> 'SSLv23:!SSLv2:!SSLv3',
 	SSL_cipher_list	=> 'ALL:!aNULL:!NULL:!LOW:!EXP:!ADH:@STRENGTH',
-	SSL_verify_mode	=> 0x01,
-	SSL_ca_path	=> '/etc/ssl/certs',
+	SSL_verify_mode	=> 0x03,
+	# Most portable to let this be discovered automatically, but
+	# an installation might override it here:
+	#SSL_ca_path	=> '/etc/ssl/certs',
 );
 # These defaults can be overriden on the cmdline:
 my ($forbid_clearauth, $forbid_clearchan) = (0, 0);
 
+# Add a key to this to blacklist that authentication mechanism.  Might be
+# useful on some platforms with broken libraries.  Make sure the key is
+# upper-case!
+my %blacklist_auth_mechanisms = ();
+# my %blacklist_auth_mechanisms = ( GSSAPI => 1, SPNEGO => 1 );
+
+# This says "go ahead and use SRV records and local hostname to figure out
+# a server to connect to".
+my $DERIVE_SIEVE_SERVER = 1;
+
+# Command used for generating local-side listings.
 my @cmd_localfs_ls = qw( ls -C );
 
+# Unset this to disable probing for SSL certs and just use whatever is
+# set above in %ssl_options.
 my $SEARCH_FOR_CERTS_DIR_IF_NEEDED = 1;
 
+# You can override this to a particular path; is only used to find default
+# certificate stores; set to undef to skip asking OpenSSL and just check
+# some common locations.
+my $OPENSSL_COMMAND = 'openssl';
+
 # ######################################################################
 # No user-serviceable parts below
 
-our $VERSION = 0;
-my $svn_revision = '$Revision: 82 $';
-if ($svn_revision =~ /^.Revision:\s*(\d+)\s*\$\z/) {
-	$svn_revision = $1;
-	$VERSION = '0.' . $1;
-} else {
-	$svn_revision = '0 because unknown';
-}
+our $VERSION = '0.87';
 
 use Authen::SASL 2.11 qw(Perl);
 # 2.11: first version with non-broken DIGEST-MD5
@@ -81,18 +92,41 @@ use Cwd qw();
 use Errno;
 use File::Basename qw();
 use File::Spec;
+use File::Temp qw/ tempfile /;
 use Getopt::Long;
 use IO::File;
 use IO::Socket::INET6;
-use IO::Socket::SSL 0.97; # SSL_ca_path bogus before 0.97
+use IO::Socket::SSL 1.14; # first version with automatic verification
 use MIME::Base64;
 use Net::DNS;
+use Net::SSLeay 1.37; # see version note below
 use Pod::Usage;
 use POSIX qw/ strftime /;
+use Sys::Hostname qw();
 use Term::ReadKey;
+
 # interactive mode will attempt to pull in Term::ReadLine too.
 
+# Net::SSLeay -- we used to just conditionally use this if loaded, indirectly
+# via the IO::Socket::SSL library.  We now want to explicitly require it for a
+# top-level feature, --tlsfingerprint.
+# The first version supplying X509_get_fingerprint() was 1.37 (2011-09-16) per
+# http://cpansearch.perl.org/src/MIKEM/Net-SSLeay-1.55/Changes
+#
+# Debian Squeeze is still on Net::SSLeay 1.36 but at this point I'm not going
+# to worry about backport compatibility to "oldstable".
+
+# This is only used to derive a default IMAP server using SRV records and
+# isn't always needed even then, so is strictly optional.
+our $have_mozilla_public_suffix;
+BEGIN { eval {
+	require 'Mozilla/PublicSuffix.pm';
+	Mozilla::PublicSuffix->import('public_suffix');
+	$have_mozilla_public_suffix = 1;
+} };
+
 my $DEBUGGING = 0;
+my $LOST_CONNECTION = 0;
 
 sub do_version_display {
 	print "${0}: Version $VERSION\n";
@@ -110,6 +144,7 @@ sub do_version_display {
 				'Authen::SASL::Perl',
 				'IO::Socket::INET6',
 				'IO::Socket::SSL',
+				'Mozilla::PublicSuffix',
 				'Net::DNS',
 				'Term::ReadKey',
 				@do_require) {
@@ -136,13 +171,15 @@ sub closedie;
 sub closedie_NOmsg;
 sub die_NOmsg;
 sub fixup_ssl_configuration;
+sub derive_sieve_server;
 
 my $DEBUGGING_SASL = 0;
 my $DATASTART = tell DATA;
 my $localsievename;
 my $remotesievename;
 my $port = undef;
-my ($user, $authzid, $authmech, $sslkeyfile, $sslcertfile, $passwordfd);
+my ($user, $authzid, $authmech, $sslkeyfile, $sslcertfile, $ssl_cert_fingerprint, $passwordfd);
+my ($tlscapath, $tlscafile);
 my $prioritise_auth_external = 0;
 my $dump_tls_information = 0;
 my $opt_version_req = 0;
@@ -153,6 +190,7 @@ my $net_domain = AF_UNSPEC;
 my $action = 'command-loop';
 my $execscript;
 GetOptions(
+	# settings which adjust how we connect
 	"localsieve=s"	=> \$localsievename,
 	"remotesieve=s"	=> \$remotesievename,
 	"server|s=s"	=> \$server,
@@ -167,6 +205,9 @@ GetOptions(
 	"clientcert=s"	=> \$sslcertfile,
 	"clientkeycert=s" => sub { $sslkeyfile = $sslcertfile = $_[1] },
 	"notlsverify|nosslverify" => sub { $ssl_options{'SSL_verify_mode'} = 0x00 },
+	"tlscertfingerprint|sslcertfingerprint=s" => \$ssl_cert_fingerprint,
+	"tlscapath=s"	=> \$tlscapath,
+	"tlscafile=s"	=> \$tlscafile,
 	"noclearauth"	=> \$forbid_clearauth,
 	"noclearchan"	=> sub { $forbid_clearauth = $forbid_clearchan = 1 },
 	"4"		=> sub { $net_domain = AF_INET },
@@ -176,6 +217,7 @@ GetOptions(
 	"dumptlsinfo|dumpsslinfo"	=> \$dump_tls_information,
 	"ignoreserverversion"	=> \$ignore_server_version,
 	# option names can be short-circuited, $action is complete:
+	# start with simple mappings to the protocol level
 	"upload"	=> sub { $action = 'upload' },
 	"download"	=> sub { $action = 'download' },
 	"list"		=> sub { $action = 'list' },
@@ -183,13 +225,22 @@ GetOptions(
 	"activate"	=> sub { $action = 'activate' },
 	"deactivate"	=> sub { $action = 'deactivate' },
 	"checkscript"   => sub { $action = 'checkscript' },
+	# then derived commands and alternate, more complex, actions
+	"edit"		=> sub { $action = 'edit' },
 	"exec|e=s"	=> sub { $execscript = $_[1]; $action='command-loop' },
+	# then administrivia
 	'help|?'	=> sub { pod2usage(0) },
 	'man'		=> sub { pod2usage(-exitstatus => 0, -verbose => 2) },
 	'version'	=> \$opt_version_req, # --version --debug should work
 ) or pod2usage(2);
 # We don't implement HAVESPACE <script> <size>
 
+if (defined $tlscafile) {
+	$ssl_options{'SSL_ca_file'} = $tlscafile;
+	delete $ssl_options{'SSL_ca_path'};
+} elsif (defined $tlscapath) {
+	$ssl_options{'SSL_ca_path'} = $tlscapath;
+}
 do_version_display() if $opt_version_req;
 
 fixup_ssl_configuration();
@@ -214,8 +265,16 @@ unless (defined $server) {
 		$server = $ENV{'IMAP_SERVER'};
 		# deal with a port number.
 		unless ($server =~ /:.*:/) { # IPv6 address literal
+			# We ignore this port because it's for the IMAP server,
+			# but we're after the Sieve server.
 			$server =~ s/:\d+\z//;
 		}
+	} elsif ($DERIVE_SIEVE_SERVER and not $no_srv) {
+		my ($tmpdomain, $tmpport, $tmpnosrv) = derive_sieve_server();
+		$server = $tmpdomain if defined $tmpdomain;
+		$port = $tmpport if defined $tmpport;
+		$no_srv = $tmpnosrv if defined $tmpnosrv;
+
 	}
 }
 
@@ -280,9 +339,15 @@ if ($action eq 'deactivate' and defined $remotesievename) {
 # ######################################################################
 # Start work; connect, start TLS, authenticate
 
+# host/port lookups are from DNS, we assume insecure. DANE might change that
+# for us in future. So, TLS negotiations do not use insecure hostnames from
+# DNS.
+my $trusted_server = $server;
+
 my @host_port_pairs;
 
 # Find the real hostname to connect to.
+# If the Sieve server was derived, then the DNS cache should be nice and warm
 unless ($no_srv) {
 	my $res = Net::DNS::Resolver->new();
 	my $query = $res->query("_sieve._tcp.$server", 'SRV');
@@ -295,8 +360,7 @@ unless ($no_srv) {
 	}
 	if (@srv_recs) {
 		@srv_recs = Net::DNS::rrsort('SRV', '', @srv_recs);
-		my $old = "[$server]:" . (defined $port ? $port : $DEFAULT_PORT);
-		debug "dns: SRV results found for: $old";
+		debug "dns: SRV results found for: $server";
 		foreach my $rr (@srv_recs) {
 			push @host_port_pairs, [$rr->target, $rr->port];
 		}
@@ -309,6 +373,7 @@ unless (@host_port_pairs) {
 	push @host_port_pairs, [$server, $port];
 }
 my $sock = undef;
+my $first_hp_attempt = 1;
 
 # Yes, this used to just try one connection and the list of candidates was
 # bolted on; how could you tell?
@@ -316,7 +381,25 @@ my $sock = undef;
 foreach my $hp (@host_port_pairs) {
 	my $host_candidate = $hp->[0];
 	my $port_candidate = $hp->[1];
-	debug "connection: trying <[$host_candidate]:$port_candidate>";
+	my $debug_host = $host_candidate =~ /:/ ? "[$host_candidate]" : $host_candidate;
+	my $debug_extra = '';
+
+	# Although we do log the actual port number, if we succeed in
+	# connecting, we don't log the actual port number tried if we don't
+	# connect because we don't have a socket to ask.  The Perl IO::Socket
+	# convention for specifying a name and fallback is not intuitively
+	# obvious and causes debugging confusion.
+	if ($port_candidate =~ /^(.+)\((\d+)\)\z/) {
+		$debug_extra .= " (try '${1}' in /etc/services, fallback $2)";
+		if ($first_hp_attempt) {
+			my @serv = getservbyname($1, 'tcp');
+			if (@serv and $serv[2] != $2) {
+				debug("connection: WARNING: /etc/services defines $1 as ${serv[2]}, not $2");
+			}
+		}
+	}
+
+	debug "connection: trying <${debug_host}:${port_candidate}>$debug_extra";
 	my $s = IO::Socket::INET6->new(
 		PeerHost	=> $host_candidate,
 		PeerPort	=> $port_candidate,
@@ -329,18 +412,20 @@ foreach my $hp (@host_port_pairs) {
 		if ($!{EINVAL} and $net_domain != AF_UNSPEC) {
 		  $extra = " (Probably no host record for overriden IP version)\n";
 		}
-		warn qq{Connection to <[$host_candidate]:$port_candidate> failed: $!\n$extra};
+		warn qq{Connection to <${debug_host}:${port_candidate}> failed: $!\n$extra};
 		next;
 	}
 	unless ($s->peerhost()) {
 		# why am I seeing successful returns for unconnected sockets? *sigh*
-		warn qq{Connection to <[$host_candidate]:$port_candidate> failed.\n};
+		warn qq{Connection to <${debug_host}:${port_candidate}> failed.\n};
 		next;
 	}
 	$sock = $s;
 	$server = $host_candidate;
 	$port = $port_candidate;
 	last;
+} continue {
+	$first_hp_attempt = 0;
 }
 exit(1) unless defined $sock;
 
@@ -506,6 +591,40 @@ parse_capabilities $sock;
 my $tls_bitlength = -1;
 
 if (exists $capa{STARTTLS}) {
+	# Always set SSL_hostname for SNI, only set the verification
+	# modes if not disabled.
+	$ssl_options{'SSL_hostname'} = $trusted_server;
+	if ($ssl_options{'SSL_verify_mode'} != 0x00) {
+		$ssl_options{'SSL_verifycn_name'} = $trusted_server;
+		# we want full wildcard support in same style that most admins
+		# are already familiar with:
+		$ssl_options{'SSL_verifycn_scheme'} = 'http';
+	}
+
+	if ($DEBUGGING) {
+
+	debug("-T- will use TLS certs from " .
+		( exists $ssl_options{'SSL_ca_file'} ? "file" : "directory" ) .
+		" \"" .
+		( exists $ssl_options{'SSL_ca_file'}
+			? $ssl_options{'SSL_ca_file'}
+			: $ssl_options{'SSL_ca_path'} ) .  "\"");
+	my $dbg_tls_verification = ' unknown';
+	if (exists $ssl_options{'SSL_verify_mode'}) {
+		my $mode = $ssl_options{'SSL_verify_mode'};
+		if ($mode == 0x00) {
+			$dbg_tls_verification = ' !DISABLED!';
+		} else {
+			$dbg_tls_verification = '';
+			$dbg_tls_verification .= ' verify-peer' if $mode & 0x01;
+			$dbg_tls_verification .= ' cert-required' if $mode & 0x02;
+			$dbg_tls_verification .= ' verify-once' if $mode & 0x04;
+		}
+	}
+	debug("-T- using hostname '${trusted_server}', verification$dbg_tls_verification");
+
+	} # DEBUGGING
+
 	ssend $sock, "STARTTLS";
 	sget $sock;
 	die "STARTTLS request rejected: $_\n" unless /^OK\b/;
@@ -517,7 +636,7 @@ if (exists $capa{STARTTLS}) {
 		my $t = Net::SSLeay::get_cipher_bits($sock->_get_ssl_object(), 0);
 		$tls_bitlength = $t if defined $t and $t;
 	}
-	debug("--- TLS activated here [$tls_bitlength bits]");
+	debug("-T- TLS activated here [$tls_bitlength bits]");
 	if ($dump_tls_information) {
 		print $sock->dump_peer_certificate();
 		if ($DEBUGGING and
@@ -529,6 +648,24 @@ if (exists $capa{STARTTLS}) {
 				$sock->peer_certificate());
 		}
 	}
+	if (defined $ssl_cert_fingerprint) {
+		my ($fp_type, $fp_want) = split(/:/, $ssl_cert_fingerprint, 2);
+		# type validation: Net::SSLeay::X509_get_fingerprint supports
+		# some set of algorithms, we don't want to overconstrain; when
+		# fed garbage, that routine silently falls back to SHA1.  So,
+		# no error return, no die to catch, we just let the resulting
+		# mismatch be shown.
+		$fp_want = uc $fp_want;
+		my $server_fingerprint = Net::SSLeay::X509_get_fingerprint(
+			$sock->peer_certificate(), $fp_type);
+		if (uc($server_fingerprint) eq $fp_want) {
+			debug("-T- TLS X.509 Fingerprint matched; [${fp_type}] $fp_want");
+		} else {
+			die "TLS X.509 Fingerprint verification failed (digest type $fp_type):\n" .
+				" expected fingerprint: $fp_want\n" .
+				"      got fingerprint: $server_fingerprint\n";
+		}
+	}
 	$forbid_clearauth = 0;
 	# The current protocol spec says that the capability response must
 	# be sent by the server after TLS is established by STARTTLS,
@@ -600,7 +737,7 @@ if (defined $realm) {
 }
 my $prompt_for_password = sub {
 	ReadMode('noecho');
-	{ print STDERR "Sieve/IMAP Password: "; $| = 1; }
+	{ print STDERR "Sieve/IMAP Password: "; local $| = 1; }
 	my $password = ReadLine(0);
 	ReadMode('normal');
 	print STDERR "\n";
@@ -629,7 +766,14 @@ if (defined $authmech) {
 	}
 	$authen_sasl_params{mechanism} = $authmech;
 } else {
-	$authen_sasl_params{mechanism} = $raw_capabilities{SASL};
+	if (scalar keys %blacklist_auth_mechanisms) {
+		$authen_sasl_params{mechanism} = join " ", grep
+			{not exists $blacklist_auth_mechanisms{uc $_}}
+			map {uc $_} @{$capa{SASL}};
+		debug("-A- Filtered mechanism list: $authen_sasl_params{mechanism}");
+	} else {
+		$authen_sasl_params{mechanism} = $raw_capabilities{SASL};
+	}
 }
 
 my $sasl = Authen::SASL->new(%authen_sasl_params);
@@ -748,6 +892,7 @@ sub sieve_delete;
 sub sieve_download;
 sub sieve_upload;
 sub sieve_checkscript;
+sub sieve_edit;
 sub localfs_ls;
 sub localfs_chpwd;
 sub localfs_pwd;
@@ -758,6 +903,8 @@ sub aux_list_keywords;
 sub complete_rl_sieve;
 sub system_result;
 sub tilde_expand ; # don't apply to cmdline params as shell does it for us
+sub determine_text_editor;
+sub prompt_retry_quit;
 
 # Do *NOT* include any sort of shell-out.
 # Basic local navigation and diagnostics yes; Yet Another ShellOut Cmd no.
@@ -857,6 +1004,15 @@ my %sieve_commands = (
 		remote_name => 1,
 		local_name => 2,
 	},
+	edit => {
+		routine => \&sieve_edit,
+		help => '<script> -- retrieve, edit, check, put script',
+		action => 1,
+		params => 1,
+		params_max => 1,
+		remote_name => 1,
+		min_version => undef, # we use checkscript, which also requires a version-compliant server
+	},
 	get	=> { alias => 'download' },
 	view	=> {
 		routine => sub { sieve_download($_[0],$_[1],'-') },
@@ -986,6 +1142,11 @@ if ($action ne 'command-loop') {
 my ($cmdlineget_func, $cmdlinedone_func);
 my $report_lineno = 0;
 
+$SIG{'PIPE'} = sub {
+	$LOST_CONNECTION = 1;
+	die "QUIT:Lost connection unexpectedly.\n";
+};
+
 if (defined $execscript) {
 	$report_lineno = 1;
 	my $scripth = new IO::File $execscript, '<'
@@ -1141,6 +1302,7 @@ while (defined (my $cmdline = $cmdlineget_func->())) {
 		}
 		last;
 	} elsif ($@) {
+		die $@ if $LOST_CONNECTION;
 		warn $@;
 	}
 }
@@ -1170,12 +1332,13 @@ sub sieve_list
 		print "$_\n";
 		sget $sock;
 	}
+	return 1;
 }
 
 sub sieve_deactivate
 {
 	my $sock = shift;
-	sieve_activate($sock, "");
+	return sieve_activate($sock, "");
 }
 
 sub sieve_activate
@@ -1186,7 +1349,9 @@ sub sieve_activate
 	sget $sock;
 	unless (/^OK((?:\s.*)?)$/) {
 		warn "SETACTIVE($scriptname) failed: $_\n";
+		return 0;
 	}
+	return 1;
 }
 
 sub sieve_delete
@@ -1197,7 +1362,9 @@ sub sieve_delete
 	sget $sock;
 	unless (/^OK((?:\s.*)?)$/) {
 		warn "DELETESCRIPT($delname) failed: $_\n";
+		return 0;
 	}
+	return 1;
 }
 
 sub sieve_download
@@ -1245,6 +1412,7 @@ sub sieve_download
 	if (defined $fh) {
 		$fh->close() or die "write-close($localfn) failed: $!\n";
 	}
+	return 1;
 }
 
 sub sieve_upload
@@ -1279,7 +1447,9 @@ sub sieve_upload
 
 	unless (/^OK((?:\s.*)?)$/) {
 		warn "PUTSCRIPT($remotefn) failed: $_\n";
+		return 0;
 	}
+	return 1;
 }
 
 sub sieve_checkscript
@@ -1306,7 +1476,45 @@ sub sieve_checkscript
 
 	unless (/^OK((?:\s.*)?)$/) {
 		warn "CHECKSCRIPT failed: $_\n";
+		return 0;
 	}
+	return 1;
+}
+
+sub sieve_edit
+{
+	my ($sock, $remotefn) = @_; splice @_, 0, 2;
+	my ($rc, $key);
+
+	my ($fh,$localfn) = tempfile();
+
+	sieve_download($sock, $remotefn, $localfn)
+		or die "failed to download script";
+
+	my @editor_cmd = determine_text_editor();
+	debug("Text editor is @editor_cmd");
+
+	while (1) {
+		system(@editor_cmd, $localfn);
+		if ($?) {
+			warn "Editor failed: $?\n";
+			$key = prompt_retry_quit();
+			return if $key eq 'q';
+			next;
+		}
+
+		sieve_checkscript($sock, $localfn) or do {
+			$key = prompt_retry_quit();
+			return if $key eq 'q';
+			next;
+		};
+		last;
+	}
+
+	print "Uploading $remotefn\n";
+	sieve_upload($sock, $localfn, $remotefn)
+		or die "failed to upload script";
+	return 1;
 }
 
 sub localfs_ls
@@ -1540,6 +1748,7 @@ sub sget
 	$l = $sock->getline() unless defined $l;
 	unless (defined $l) {
 		debug "... no line read, connection dropped?";
+		$LOST_CONNECTION = 1;
 		die "Connection dropped unexpectedly when trying to read.\n";
 	}
 	if ($l =~ /{(\d+)\+?}\s*\n?\z/) {
@@ -1553,6 +1762,13 @@ sub sget
 				$len -= length($extra);
 				$l .= $extra;
 			}
+			# there's a CRLF _after_ the literal string
+			if ($len == 0) {
+				my $hope_crlf = $sock->getline();
+				if ($hope_crlf ne "\r\n") {
+					debug "... after literal, did not get final CRLF but \"${hope_crlf}\"";
+				}
+			}
 		}
 		$dochomp = 0;
 	}
@@ -1570,7 +1786,7 @@ sub sget
 sub sfinish
 {
 	my $sock = shift;
-	if (defined $_[0]) {
+	if (defined $_[0] and not $LOST_CONNECTION) {
 		ssend $sock, $_[0];
 		sget $sock;
 	}
@@ -1642,23 +1858,265 @@ sub tilde_expand
 	return ($more and wantarray) ? ($path, $tilded, $home) : $path;
 }
 
-sub fixup_ssl_configuration {
+# Given a directory, the fact of its existence doesn't mean it's usable for
+# OpenSSL certs; it could have been created by an installer but never used.  If
+# it's usable, then there will be hash symlinks therein.  Each is 8 hex chars,
+# a dot and then a sequence number starting at 0, to handle hash collisions.
+# So, if there are no (symlinks to) files matching *.0 then although the dir
+# exists, it's unitialised cruft.  If none of the symlinks are resolvable, it's
+# also cruft.
+#
+# Theoretically, a system might try to make the dir unreadable-but-executable
+# so that it becomes an oracle that can be asked if a given cert should be
+# trusted.  That approach of self-mistrust is not supported: if you have such a
+# system, report it and we can add an option to disable the contents probe.
+sub confirm_valid_sslcerts_dir
+{
+	my $dir = shift;
+	die "internal error" unless defined $dir;
+	return 0 unless -e $dir;
+	unless (-d $dir) {
+		debug("setup: not a directory: $dir");
+		return 0;
+	}
+	unless (opendir(DIR, $dir)) {
+		debug("setup: unable to opendir($dir): $!");
+		return 0;
+	}
+	my @some_entries = grep /^[^.].+\.0\z/, readdir(DIR);
+	# while opendir failing is likely permissions, closedir
+	# failing is the system failing badly.
+	closedir(DIR) or die "closedir($dir) failed: $!\n";
+	foreach my $entry (@some_entries) {
+		return 1 if -f File::Spec->catfile($dir, $entry);
+	}
+	debug("setup: found no files named for cert-hashes, rejecting dir [$dir]");
+	return 0;
+}
+
+sub fixup_ssl_configuration
+{
 	return unless $SEARCH_FOR_CERTS_DIR_IF_NEEDED;
-	return if -d $ssl_options{'SSL_ca_path'};
-	debug "Need to find SSL_ca_path, trying to ask openssl";
-	my $found = 0;
+	return if exists $ssl_options{'SSL_ca_file'} and -f $ssl_options{'SSL_ca_file'};
+	return if exists $ssl_options{'SSL_ca_path'} and -d $ssl_options{'SSL_ca_path'};
 	local *_;
-	open(VERSION, '-|', 'openssl', 'version', '-d');
-	foreach (<VERSION>) {
-		next unless /^OPENSSLDIR: "(.+)"\s*$/;
-		$ssl_options{'SSL_ca_path'} = File::Spec->catdir($1, 'certs');
-		$found = 1;
-		last;
+	delete @ssl_options{'SSL_ca_path', 'SSL_ca_file'};
+
+	# OpenSSL crypto/cryptlib.h defines env var names
+	# SSL_CERT_DIR & SSL_CERT_FILE; these are expected to take precedence
+	# over any defaults from the library.  If the library is asked to load
+	# defaults, it will honour these; unfortunately, IO::Socket::SSL
+	# attempts to provide defaults at its own layer and neither it nor
+	# Net::SSLeay has a facility to ask that OpenSSL defaults be loaded.
+	# Thus we have to emulate support to get back to something reasonable.
+	if (exists $ENV{'SSL_CERT_DIR'}) {
+		if (confirm_valid_sslcerts_dir($ENV{'SSL_CERT_DIR'})) {
+			$ssl_options{'SSL_ca_path'} = $ENV{'SSL_CERT_DIR'};
+			debug("setup: Have set SSL_ca_path to $ssl_options{'SSL_ca_path'} (from \$SSL_CERT_DIR)");
+			return;
+		}
+		debug("setup: \$SSL_CERT_DIR defined but invalid, ignoring")
+	}
+	if (exists $ENV{'SSL_CERT_FILE'}) {
+		if (-f $ENV{'SSL_CERT_FILE'}) {
+			$ssl_options{'SSL_ca_file'} = $ENV{'SSL_CERT_FILE'};
+			debug("setup: Have set SSL_ca_file to $ssl_options{'SSL_ca_file'} (from \$SSL_CERT_FILE)");
+			return;
+		}
+	}
+
+	if (defined $OPENSSL_COMMAND) {
+		debug "setup: Need to find SSL_ca_path, trying to ask openssl";
+		my $found = undef;
+		# protect against openssl command not existing
+		open (my $olderr, ">&STDERR") or die "failed to dup(stderr): $!\n";
+		open(STDERR, File::Spec->devnull());
+		my $oops = 0;
+		if (open(VERSION, '-|', $OPENSSL_COMMAND, 'version', '-d')) {
+			foreach (<VERSION>) {
+				next unless /^OPENSSLDIR: "(.+)"\s*$/;
+				$found = $1;
+				last;
+			}
+			close(VERSION);
+		} else {
+			$oops = 1;
+		}
+		open(STDERR, ">&", $olderr); close($olderr);
+		if ($oops) {
+			debug("setup: unable to run openssl");
+		} elsif (defined $found) {
+			my $attempt = File::Spec->catdir($found, 'certs');
+			if (confirm_valid_sslcerts_dir($attempt)) {
+				$ssl_options{'SSL_ca_path'} = $attempt;
+				debug("setup: Have set SSL_ca_path to $ssl_options{'SSL_ca_path'}");
+			} else {
+				debug("setup: found OPENSSLDIR but certs/ invalid");
+				$found = undef;
+			}
+		} else {
+			debug("setup: openssl did not tell us OPENSSLDIR");
+		}
+		return if defined $found;
+	}
+
+	debug "setup: No OpenSSL, check some common locations";
+	# ripped from $GOROOT/src/pkg/crypto/x509/root_unix.go :
+	my @golang_locations = (
+		"/etc/ssl/certs/ca-certificates.crt",     # Linux etc
+		"/etc/pki/tls/certs/ca-bundle.crt",       # Fedora/RHEL
+		"/etc/ssl/ca-bundle.pem",                 # OpenSUSE
+		"/etc/ssl/cert.pem",                      # OpenBSD
+		"/usr/local/share/certs/ca-root-nss.crt", # FreeBSD
+	);
+	foreach my $loc (@golang_locations) {
+		-f $loc or next;
+		$ssl_options{'SSL_ca_file'} = $loc;
+		debug("setup: Have set SSL_ca_file to $ssl_options{'SSL_ca_file'}");
+		return;
+	}
+	my @alt_dir_locations = (
+		"/etc/ssl/certs",			# widespread
+		"/etc/x509/certs",
+		"/system/etc/security/cacerts",		# some Android
+		"/data/misc/keychain/cacerts-added",	# some Android
+	);
+	foreach my $loc (@alt_dir_locations) {
+		confirm_valid_sslcerts_dir($loc) or next;
+		$ssl_options{'SSL_ca_path'} = $loc;
+		debug("setup: Have set SSL_ca_path to $ssl_options{'SSL_ca_path'}");
+		return;
+	}
+}
+
+# returns 1 "okay", 0 "definitely not available", undef "unknown"
+sub domain_sieve_server
+{
+	my $domain = shift;
+	my $res = Net::DNS::Resolver->new();
+	foreach my $protocol ('sieve', 'imaps', 'imap') {
+		my $srv_name = "_${protocol}._tcp.${domain}";
+		my $query = $res->query($srv_name, 'SRV');
+		next unless $query;
+		my $okay = undef;
+		# if we have N>1 records and 1 says "no service present", it
+		# wins always, as the alternative is non-deterministic. Given
+		# buggy DNS entries, it's better to fail reproducibly always
+		# instead of only failing intermittently.
+		foreach my $rr ($query->answer) {
+			next unless $rr->type eq 'SRV';
+			if ($rr->target eq '.') {
+				$okay = 0;
+				last;
+			}
+			$okay = 1;
+		}
+		if (defined $okay) {
+			debug "findserver: found SRV record '${srv_name}' is @{[$okay ? '' : 'NOT ']} available";
+			return $okay;
+		}
+	}
+	return undef;
+}
+
+# returns three fields:
+#   * host/domain
+#   * port
+#   * no_srv new state
+# any may be undef to mean "no change".
+#
+# If we find SRV records, which state Sieve/IMAP present, we return the domain
+# and explicitly state no_srv off so that later we can correctly iterate
+# through.  If they say "explicitly not present", we fail to match.
+#
+# If the Mozilla::PublicSuffix module is present, we can walk up the hostname
+# until public, otherwise we only check one level up.
+#
+# Limitation: assumes domain-names well-formed for use as hostnames, doesn't
+# handled embedded dots within a label.
+sub derive_sieve_server
+{
+	return (undef, undef, undef) unless $DERIVE_SIEVE_SERVER;
+	my $host = Sys::Hostname::hostname();
+
+	# Some people actually do use a hostname that is their domain, where
+	# the domain is a public suffix. We won't support that unless we can
+	# be _sure_ that we're not checking public domains, by having
+	# Mozilla::PublicSuffix installed.
+	#
+	# Note that example.co.uk as a hostname would still check co.uk --
+	# for bug-free operation with hosts that are the domain and the domain
+	# is public, use that module!
+	my $domain;
+	if ($host =~ /^[^.]+\.(.+?\..+)\z/) {
+		$domain = $1;
+	} else {
+		return (undef, undef, undef) unless $have_mozilla_public_suffix;
+	}
+
+	if ($have_mozilla_public_suffix) {
+		$domain = $host;
+		debug "findserver: walking up hostname domains with Mozilla PSL";
+		while ($domain ne public_suffix($domain)) {
+			debug "findserver: checking SRV records for: $domain";
+			my $have_srv = domain_sieve_server $domain;
+			if (defined $have_srv) {
+				if ($have_srv) {
+					debug "findserver: SRV records, match";
+					return ($domain, undef, 0);
+				}
+				debug "findserver: SRV record explicitly says 'no service'";
+				return (undef, undef, 1); # force off SRV lookups
+			}
+			if ($domain =~ /^[^.]+\.(.+)\z/) {
+				$domain = $1;
+			} else {
+				last;
+			}
+		}
+	} else {
+		debug "findserver: checking SRV records for: $domain";
+		my $have_srv = domain_sieve_server $domain;
+		if (defined $have_srv) {
+			if ($have_srv) {
+				debug "findserver: SRV records, match";
+				return ($domain, undef, 0);
+			}
+			debug "findserver: SRV record explicitly says 'no service'";
+			return (undef, undef, 1); # force off SRV lookups
+		}
+	}
+
+	# We give up, say nothing
+	debug "findserver: no clues found";
+	return (undef, undef, undef);
+}
+
+sub determine_text_editor {
+	foreach my $envvarname (qw(VISUAL EDITOR)) {
+		if (exists $ENV{$envvarname} and $ENV{$envvarname} ne "") {
+			return split(/\s+/, $ENV{$envvarname});
+		}
+	}
+	foreach my $cmd (qw(sensible-editor vi)) {
+		foreach my $p (File::Spec->path()) {
+			my $candidate = File::Spec->catfile($p, $cmd);
+			return ($candidate) if -x $candidate;
+		}
+	}
+}
+
+sub prompt_retry_quit {
+	while (1) {
+		print STDERR "(R)edit or (Q)uit: ";
+		local $| = 1;
+		my $key = ReadKey(0);
+		print STDERR "\n";
+		next unless defined $key;
+		chomp $key;
+		$key = lc($key);
+		return $key if $key eq 'q' or $key eq 'r';
 	}
-	close(VERSION);
-	debug($found
-		? "Have set SSL_ca_path to $ssl_options{'SSL_ca_path'}"
-		: "Unable to get system SSL_ca_path");
 }
 
 # ######################################################################
@@ -1672,20 +2130,22 @@ sieve-connect - managesieve command-line client
 
  sieve-connect [-s <hostname>] [-p <portspec>] [-u <user>] [a <authzid>]
                [-m <authmech>] [-r realm] [-e execscript]
-	       [... longopts ...]
+               [... longopts ...]
  sieve-connect [--localsieve <script>] [--remotesieve <script>]
-	       [--debug] [--dumptlsinfo]
+               [--debug] [--dumptlsinfo]
                [--server <hostname>] [--port <portspec>] [--4|--6]
-	       [--user <authentication_id>] [--authzid <authzid>]
-	       [--realm <realm>] [--passwordfd <n>]
-	       [--clientkey <file> --clientcert <file>]|[--clientkeycert <file>]
-	       [--notlsverify|--nosslverify]
-	       [--noclearauth] [--noclearchan]
-	       [--authmech <mechanism>]
-	       [--ignoreserverversion]
-	       [--upload|--download|--list|--delete|--check
-	        --activate|--deactivate]|[--exec <script>]
-	       [--help|--man]
+               [--user <authentication_id>] [--authzid <authzid>]
+               [--realm <realm>] [--passwordfd <n>]
+               [--clientkey <file> --clientcert <file>]|[--clientkeycert <file>]
+               [--notlsverify|--nosslverify]
+               [--tlscertfingerprint|--sslcertfingerprint <dgsttype:digest>]
+               [--tlscapath <ca_directory>]|[--tlscafile <ca_file>]
+               [--noclearauth] [--noclearchan]
+               [--authmech <mechanism>]
+               [--ignoreserverversion]
+               [--upload|--download|--list|--delete|--checkscript|--edit|
+                --activate|--deactivate]|[--exec <script>]
+               [--help|--man]
 
 =head1 DESCRIPTION
 
@@ -1716,6 +2176,8 @@ etc.
 
 =head1 OPTIONS
 
+Option names may be given as the shortest unique prefix.
+
 The remote sieve script name defaults to the same as the local sieve
 script name, so just specify the local one if only one is needed; it
 was a deliberate decision to have the defaults this way around, to make
@@ -1730,12 +2192,35 @@ provided as debug trace.
 
 The B<--version> option shows version information.
 When combined with B<--debug> it will show implementation dependency versions.
+The B<--help> and B<--man> options provide usage information.
+
+The server can be a host or IP address, IPv4 or IPv6.
+
+If a server is provided by B<--server> then that takes precedence.
+If that option is not present, then C<$IMAP_SERVER> from the environment is
+checked and, if it's not a unix-domain socket path, is used with any port
+specification stripped off.
+
+Next, unless B<--nosrv> is given, checks are made for SRV records so as to search
+for a default server; if the F<Mozilla::PublicSuffix> Perl module is available,
+these checks are done for every level of the hostname upto (but not including)
+the public suffix.
+If that module is not available, a crude heuristic is used: as long as there
+are three dots in the hostname, SRV records for the part of the hostname after
+the first dot are tried.
+If this is inappropriate, install F<Mozilla::PublicSuffix>.
+
+If no SRV records are found which point to a 'sieve', 'imaps' or 'imap'
+protocol service, of if a record is found which says "no such service in
+this domain" (by having a target of "."), then the final default server
+is F<localhost>.
 
-The server can be a host or IP address, IPv4 or IPv6;
-the default is C<$IMAP_SERVER> from the environment (if it's not a
-unix-domain socket path) with any port specificaion stripped off,
-else F<localhost>.
 The port can be any Perl port specification, default is F<sieve(4190)>.
+A port from an SRV record will take precedence.  The Perl specification
+provides a name to look up in the system services database (F</etc/services>)
+followed in parentheses by a default value to use if the name is not found.
+Thus this default will honour a value of 2000 from F</etc/services>.
+
 The B<--4> or B<--6> options may be used to coerce IPv4 or IPv6.
 
 By default, the server is taken to be a domain, for which SRV records are
@@ -1755,11 +2240,38 @@ Everything until the newline before EOF is the password,
 so it can contain embedded newlines.  Do not provide passwords on a
 command-line or in a process environment.
 
-If you are willing to accept the risk of man-in-the-middle active attacks
-and you are unable to arrange for the relevant Certificate Authority
-certificate to be available, then you can lower your safety with the
+Unless modified at install/packaging time, by default SSL certificate authority
+certificates are searched for.
+The first attempt is to try, in turn, for environment variables
+C<$SSL_CERT_DIR> & C<$SSL_CERT_FILE> which are the names supported by the
+OpenSSL library and so often supported by client commands.
+Next, if the OpenSSL command C<version> is available and the output
+C<OPENSSLDIR> can be parsed and the C<certs> directory exists within that
+directory, then that location will be used.
+Finally, a fixed list of common locations are searched and the first one to
+exist is used.  Invoking with B<--debug> will show more details during the
+C<setup:> phase.
+
+Precedence above these defaults is given to the B<--tlscafile> option if
+given, else the B<--tlscapath> option if that is given.
+The former is one file containing certificates, the latter is a directory.
+
+Alternatively, if you are willing to accept the risk of man-in-the-middle
+active attacks and you are unable to arrange for the relevant Certificate
+Authority certificate to be available, then you can lower your safety with the
 B<--notlsverify> option, also spelt B<--nosslverify>.
 
+If you don't want to (only) rely on CA systems you can explicitly set an
+expected server certificate fingerprint using the B<--tlscertfingerprint>
+option, also spelt B<--sslcertfingerprint>.  If you wish to ignore CA
+validation, you still need to disable that explicitly (see above), as the
+default is to add an extra constraint (pinning, within valid CA certificates).
+This option specifies the X.509 certificate fingerprint (not a public key
+fingerprint), as given by OpenSSL.  The first part of the value should be an
+algorithm name, such as C<sha256> or C<sha1>.  That is followed by a colon, and
+then the fingerprint data in its usual colon-delimited hexadecimal notation.
+Eg: C<--tlscertfingerprint sha256:24:B4:..28-more-fields..:A8:58>
+
 For SSL client certificate authentication, either B<--clientkeycert> may
 be used to refer to a file with both the key and cert present or both
 B<--clientkey> and B<--clientcert> should point to the relevant files.
@@ -1778,17 +2290,58 @@ present.  If no action is present, the interactive mode is entered.
 If the exec action is present, commands are read from the script
 instead.
 
-It is believed that the names of the actions are
-sufficiently self-descriptive for any English-speaker who can safely be
-allowed unaccompanied computer usage.
+=over 4
+
+=item B<--upload>
+will upload a script to the server.
+
+=item B<--download>
+will download a script from the server.
+
+=item B<--list>
+will list the scripts which exist on the server.
+One of those scripts might be marked ACTIVE.
+
+=item B<--delete>
+will delete a script from the server.
+
+=item B<--checkscript>
+will ask the server to validate the local file provided.
+
+=item B<--edit>
+will download a script, invoke an editor upon it, ask the server to
+check the results (and offer to re-edit if the server rejects it) and finally
+upload the result.
+
+=item B<--activate>
+will mark the specified remote script as the active one.
+
+=item B<--deactivate>
+will remove the active mark from the specified remote script
+without activating a replacement.
+
+=item B<--exec>
+will take a file-name containing commands as though given in the normal
+read-eval-print loop.
+
+=back
+
+Note that B<--check> and B<--edit> require a server which advertises
+a "VERSION" capability, see B<--ignoreserverversion> to override.
 
 (If B<--server> is not explicitly stated, it may be provided at the end of
 the command-line for compatibility with sieveshell.)
 
 =head1 ENVIRONMENT
 
-C<$IMAP_SERVER> for a default IMAP server.  C<$USERNAME> and C<$LOGNAME>
-where the C<getpwuid()> function is not available.
+C<$IMAP_SERVER> for a default IMAP server.
+
+C<$USERNAME> and C<$LOGNAME> where the C<getpwuid()> function is not available.
+
+C<$SSL_CERT_DIR> and C<$SSL_CERT_FILE> for locating default
+Certificate Authority trust anchors.
+
+C<$VISUAL>, else C<$EDITOR>, for the edit action.
 
 =head1 BUGS
 
@@ -1809,10 +2362,10 @@ issue in the F<IO::Socket::INET6> Perl library and need to upgrade that.
 
 Most historical implementations used port 2000 for ManageSieve.  RFC5804
 allocates port 4190.  This tool uses a port-spec of "sieve(4190)" as the
-default port, which means that an /etc/services (or substitute) entry for
-"sieve" as a TCP service takes precedence, but if that is not present, to
+default port, which means that an F</etc/services> (or substitute) entry for
+"sieve" as a TCP service takes precedence, but if that is not present, will
 assume 4190 as the default.  This change means that if you're still using
-port 2000 and do not have an /etc/services entry, updating to/beyond release
+port 2000 and do not have an F</etc/services> entry, updating to/beyond release
 0.75 of this tool will break invocations which do not specify a port.  The
 specification of the default port was moved to the user-configurable section
 at the top of the script and administrators may wish to override the shipped
@@ -1824,6 +2377,12 @@ weighted prioritised SRV records and I have not made any effort to fix this;
 whatever the default sort algorithm provides for SRV is what is used for
 ordering.
 
+If you don't specify a server and don't export C<$IMAP_SERVER> in the
+environment then the search mechanism is safer and more thorough if the
+F<Mozilla::PublicSuffix> Perl module is installed. In particular, if your
+hostname is also your domain name and the parent domain is administered
+by someone you don't trust, then you'll regret not installing that module.
+
 Probably need to sit down and work through the final RFC and see if any
 functionality is still missing.
 
@@ -1840,6 +2399,10 @@ mailing-list, 2006-11-14.  It was a single-action-and-quit script for
 scripting purposes.  The command-loop code was written (two days) later
 and deliberately designed to be compatible with sieveshell.
 
+Versions prior to 0.85 did not actually verify the peer certificate identity,
+although this author stupidly believed that it did.
+API/expectations mismatch.
+
 =head1 AUTHOR
 
 Phil Pennock E<lt>phil-perl@spodhuis.orgE<gt> is guilty, m'Lud.
@@ -1852,13 +2415,14 @@ L<mailto:sieve-connect-announce-request@spodhuis.org?subject=subscribe>
 =head1 PREREQUISITES
 
 Perl.  F<Authen::SASL>.  F<IO::Socket::INET6>.
-F<IO::Socket::SSL> (at least version 0.97).  F<Pod::Usage>.
+F<IO::Socket::SSL> (at least version 1.14).  F<Pod::Usage>.
 F<Net::DNS> for SRV lookup.
 F<Pod::Simple::Text> for built-in man command (optional).
 F<Term::ReadKey> to get passwords without echo.
 Various other Perl modules which are believed to be standard.
 F<Term::ReadLine> will significantly improve interactive mode.
 F<Term::ReadLine::Gnu> will improve it further by allowing tab-completion.
+F<Mozilla::PublicSuffix> is highly recommended and will improve security.
 
 =head1 INTEROPERABILITY
 
